#include "WProgram.h"
#include "string_to_hex.h"

#include <string.h>
#include <stdlib.h>

#define BYTE_STRING_LENGTH      5
#define ADDRESS_LENGTH          16
#define FRAME_ID_NO_RESPONSE    0x00
#define UNKNOWN_DEST_ADD_BYTE_1 0xFF
#define UNKNOWN_DEST_ADD_BYTE_2 0xFE

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

   SET 64 BIT ADDRESS

   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
struct XBee
{
   int sixty_four_bit_address[ ADDRESS_LENGTH / 2 ];
};

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

   SET 64 BIT ADDRESS

   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
void inline set_64_bit_address_on( struct XBee xb, const char *name )
{
   if( strlen( name ) != ADDRESS_LENGTH )
   {
      Serial.println( "Invalid 64-bit address length.  Exiting...\n" );
      exit( -1 );
   }

   char *temp = ( char * ) malloc( sizeof( int ) * 3 );

   for( int i = 0, j = 0; i < ADDRESS_LENGTH; i += 2, j++ )
   {
      temp[ 0 ] = name[ i ];
      temp[ 1 ] = name[ i + 1 ];
      temp[ 2 ] = '\0';

      xb.sixty_four_bit_address[ j ] = convert_two_char_string_to_hex( temp );
   }

   free( temp );
   temp = NULL;
}

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::



   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
static void convert_string_address_to_int_address( char from[], int to[] )
{
   char *temp = ( char * ) malloc( sizeof( int ) * 3 );

   for( int i = 0, j = 0; i < ADDRESS_LENGTH; i += 2, j++ )
   {
      temp[ 0 ] = from[ i ];
      temp[ 1 ] = from[ i + 1 ];
      temp[ 2 ] = '\0';

      to[ j ] = convert_two_char_string_to_hex( temp );
   }

   free( temp );
   temp = NULL;
}

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

   FILL_API_PACKET

   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
void fill_api_packet( byte *p, int size )
{
   for( int i = 1; i < size; i++ )
      p[ i ] = Serial.read();
}

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

   PRINT_API_PACKET

   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
void print_api_packet( byte *packet, int size )
{
   if( size == 0 )
      return;
   else
   {
      Serial.print( "0x" );
      Serial.print( packet[ 0 ], HEX );

      for( int i = 1; i < size; i++ )
      {
         Serial.print( " + 0x" );
         Serial.print( packet[ i ], HEX );
      }
   }
   
   Serial.print( "\r\n" );
}

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

   GET SERIAL DATA FROM

   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
inline int get_serial_data_from( byte *a, const int PACKET_SIZE )
{
   return ( int ) a[ PACKET_SIZE - 3 ] + ( int ) a[ PACKET_SIZE - 2 ];
}


/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

   CONFIRM CHECKSUM

   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
int checksum_confirmed_in( byte *p, const int PACKET_SIZE, byte xbee_checksum )
{
   static byte calculated_checksum;
   static byte verified_checksum;

   static int i;
   static byte data_between_length_and_last;

   for( i = 0, data_between_length_and_last = 0; i < PACKET_SIZE; i++ )
      data_between_length_and_last += p[ i ];

   calculated_checksum = 0xFF & ( 0xFF - data_between_length_and_last );
   verified_checksum =   0xFF & ( data_between_length_and_last + xbee_checksum );

   return ( verified_checksum == CHECKSUM ) ? VALID : INVALID;
}

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
 
   SEND REMOTE AT QUERY
 
   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
void send_remote_at_query( struct XBee xb, char char_dest_address[], char at_command[] )
{
   int int_dest_address[ ADDRESS_LENGTH / 2 ];

   convert_string_address_to_int_address( char_dest_address, int_dest_address );

   if( strlen( at_command ) != 2 )
   {
      Serial.println( "The AT command argument to function \"remote_"
                      "at_query\" must be two characters.  Exiting..." );
      exit( -1 );
   }

   int full_sum = 0;
   int checksum = 0;

   Serial.print( START_DELIMITER, BYTE );                  // 0
   Serial.print( LENGTH_MSB, BYTE );                       // 1
   Serial.print( FIFTEEN, BYTE );                          // 2
   Serial.print( API_FRAME_REMOTE_COMMAND_REQUEST, BYTE ); // 3
   Serial.print( FRAME_ID_RESPONSE_YES, BYTE );            // 4

   for( int i = 0; i < 8; i++ )
      Serial.print( int_dest_address[ i ], BYTE );

   Serial.print( UNKNOWN_DEST_ADD_BYTE_1, BYTE );          // 13
   Serial.print( UNKNOWN_DEST_ADD_BYTE_2, BYTE );          // 14. 
   Serial.print( 0x02, BYTE );                             // 0x02 ???

   Serial.print( at_command[ 0 ], BYTE );                  // 16. the first char of the command
   Serial.print( at_command[ 1 ], BYTE );                  // 17. the second char of the command

   full_sum = 
      API_FRAME_REMOTE_COMMAND_REQUEST +
      FRAME_ID_RESPONSE_YES +
      ( byte )int_dest_address[ 0 ] + ( byte )int_dest_address[ 1 ] + 
      ( byte )int_dest_address[ 2 ] + ( byte )int_dest_address[ 3 ] + 
      ( byte )int_dest_address[ 4 ] + ( byte )int_dest_address[ 5 ] + 
      ( byte )int_dest_address[ 6 ] + ( byte )int_dest_address[ 7 ] + 
	   UNKNOWN_DEST_ADD_BYTE_1 + UNKNOWN_DEST_ADD_BYTE_2 +
	   0x02 +
	   at_command[ 0 ] + at_command[ 1 ];

   checksum = 0xFF - ( 0xFF & full_sum );

   Serial.print( checksum, BYTE );
}

/* :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

   READ AT COMMAND RESPONSE

   :::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: */
void read_at_query_into( byte *returned_data )
{
   static byte at_command_response[ 4 ];

   int tries = 0;
   int my_time = 1000;

start:
   if( tries == 5 )
      return;

   while( my_time-- )
      if( Serial.available() > 2 )
         break;

   if( ( at_command_response[ 0 ] = Serial.read() ) == START_DELIMITER )
   {
      at_command_response[ 1 ] = Serial.read();
      at_command_response[ 2 ] = Serial.read();

      byte *data = 
         ( byte * ) 
         malloc( sizeof( byte ) * (( int )at_command_response[ 2 ]) );

      for( int i = 0; i < ( int )at_command_response[ 2 ]; i++ )
         data[ i ] = Serial.read();

      at_command_response[ 3 ] = Serial.read();

      if( checksum_confirmed_in( data, ( int )at_command_response[ 2 ], at_command_response[ 3 ] ) == VALID )
      {
         *returned_data = data[ ( ( int )at_command_response[ 2 ] - 1 ) ];
         
         free( data );
         data = NULL;

         Serial.flush();
      }
      else
      {
         free( data );
         data = NULL;
         
         return;
      }
   }
   else
   {
      tries++;
      goto start;
   }

   delay( ONE_SECOND / 2 );
}
